-- Script para la actualizacion de una BBDD de ClearLight
-- a la version 6.8

-- Creacion de la tabla de Cheques devueltos
CREATE TABLE dbo.ChequesDevueltos(
	id int IDENTITY(1,1) PRIMARY KEY NOT NULL,
	CodigoCliente nvarchar(20) NOT NULL,
	Descripcion nvarchar(50) NOT NULL,
	FechaDevolucion datetime NOT NULL,
	CodigoEmisor nvarchar(20) NOT NULL,
	NumeroDocumento nvarchar(20) NOT NULL,
	Monto money NOT NULL,
	MontoPenalizacion money NOT NULL,
	Saldo money NOT NULL
)
GO

CREATE INDEX ix_ChequesDevueltosCliente ON ChequesDevueltos (CodigoCliente, FechaDevolucion)
GO

CREATE TABLE dbo.BalancesMensuales(
	TipoEntidad nvarchar(3) NOT NULL,
	CodigoEntidad nvarchar(20) NOT NULL,
	IndicePeriodo int NOT NULL,
	Balance money NOT NULL,
 CONSTRAINT PK_BalancesMensuales PRIMARY KEY CLUSTERED 
	(
		TipoEntidad ASC,
		CodigoEntidad ASC,
		IndicePeriodo ASC
	)
)
GO

-- ===================================================================================================================
-- La Tabla de BalancesMensuales mantiene los balances de las entidades mes por mes expresados en MONEDA ACTUAL.
-- si no se us la reconversin monetaria, se debe asignar el valor 0 a la variable @doReconversion
DECLARE @doReconversion AS TINYINT
SET @doReconversion = 1

INSERT INTO BalancesMensuales (TipoEntidad, CodigoEntidad, IndicePeriodo, Balance)
	SELECT TipoEntidad, CodigoEntidad, Year(Fecha) * 100 + Month(Fecha), SUM(Debe - Haber)
	FROM DetallesCuentaEntidad WHERE Not TipoEntidad In ('INV', 'VIV')
	GROUP BY TipoEntidad, CodigoEntidad, Year(Fecha) * 100 + Month(Fecha)
	
INSERT INTO BalancesMensuales (TipoEntidad, CodigoEntidad, IndicePeriodo, Balance)
	SELECT 'VIV', CodigoItem, YEAR(FechaOperacion) * 100 + MONTH(FechaOperacion), SUM((Entradas - Salidas) * Costo)
	FROM DetallesMinv
	GROUP BY CodigoItem, YEAR(FechaOperacion) * 100 + MONTH(FechaOperacion)

IF @doReconversion = 1 BEGIN
	UPDATE BalancesMensuales
	SET Balance = Balance / 1000
	WHERE IndicePeriodo < 200801
END

-- Las existencias no se reconvierten en ningun caso:
INSERT INTO BalancesMensuales (TipoEntidad, CodigoEntidad, IndicePeriodo, Balance)
	SELECT 'INV', CodigoItem, YEAR(FEchaOperacion) * 100 + MONTH(FechaOperacion), SUM(Entradas - Salidas)
	FROM DetallesMinv
	GROUP BY CodigoItem, YEAR(FechaOperacion) * 100 + MONTH(FechaOperacion)
GO


-- ===================================================================================================================
-- Las consultas SQL que usen divisiones cuyo divisor pueda ser CERO, deberan usar dbo.SafeDivide(dividendo, divisor)
-- para evitar los errores causados por las divisiones por cero
-- Es decir, en vez de:
-- 			SELECT SUM(Precio * Cantidad) / Sum(Cantidad)
-- se podr escribir 
--			SELECT dbo.SafeDivide(SUM(Precio * Cantidad), SUM(Cantidad))
-- para que en caso de que la suma decantidades sea igual a cero el resultado sea CERO en vez de producirse un erro
CREATE FUNCTION dbo.SafeDivide(@dividendo FLOAT, @divisor FLOAT) 
RETURNS FLOAT
AS
BEGIN
	IF @divisor = 0
	BEGIN
		RETURN 0
	END
	RETURN @Dividendo / @Divisor
END
GO


-- ===================================================================================================================
-- GetDocumentosMedioPago devuelve una tabla con informacion sobre los documentos afectados por un medio de pago
-- recibido. Se usa desde el componente de transcripcion de cheques devueltos
CREATE FUNCTION dbo.GetDocumentosMedioPago (
	@codigoMedio NVARCHAR(12),
  @Emisor NVARCHAR(20),
  @NumeroCheque NVARCHAR(20)
) RETURNS @DocumentosAfectados TABLE (
	Tipo nVarChar(3),
  CodigoCliente nVarchar(20),
  Referencia nVarchar(40),
  FechaTransaccion DateTime, 
	MontoCheque Money
)
AS 
BEGIN
  INSERT @DocumentosAfectados
  Select * From
  (Select 'FCT' As Tipo, CodigoCliente, Serie + '-' + RIGHT('00000000' + CONVERT(NVARCHAR(8), Correlativo), 8) As Referencia, FechaTransaccion, DET.Monto
   FROM Facturas, DetallesIngresoCaja DET
   WHERE DET.Medio = @CodigoMedio AND DET.Emisor = @Emisor AND Det.NumeroDocumento = @NumeroCheque
   AND Facturas.Numero IN (
     SELECT NumDoc
	 FROM SubDocsMC
     WHERE TipoDoc = 'FCT' 
	 AND TransID = DET.TransID)
   UNION SELECT 'NDC' As Tipo, CodigoEntidad, 'NDC-' + RIGHT('00000000' + CONVERT(NVARCHAR(8), Correlativo), 8), FechaDoc, DET2.Monto
   FROM NotasDCCP, DetallesINgresoCaja Det2
   WHERE TipoEntidad = 'CLT' 
   AND DET2.Medio = @CodigoMedio AND DET2.Emisor = @Emisor AND Det2.NumeroDocumento = @NumeroCheque
   AND Numero IN (
     SELECT NumDoc FROM SubDocsMC
     WHERE TipoDoc = 'NDC' 
	   AND TransID = DET2.TransID)
   UNION SELECT 'CHD' As Tipo, CodigoCliente, 'CHD' + RIGHT('00000000' + CONVERT(NVARCHAR(8), CH.ID), 8), FechaDevolucion, DET3.Monto
   FROM ChequesDevueltos CH, DetallesIngresoCaja Det3
   WHERE Det3.Medio = @CodigoMedio AND Det3.Emisor = @Emisor AND Det3.NumeroDocumento = @NumeroCheque
   AND CH.ID IN (
	  SELECT NumDoc FROM SubDOcsMC
	  WHERE TipoDoc = 'CHD'
      AND TransID = DET3.TransID)
  ) T
  RETURN
END
GO




-- ===================================================================================================================
-- Devuelve el saldo de una entidad para una fecha dada.
CREATE FUNCTION dbo.SaldoEntidadFecha(@TipoEntidad NVARCHAR(3), @CodigoEntidad NVARCHAR(40), @Fecha DATETIME)
RETURNS MONEY
AS
BEGIN
	DECLARE @Result MONEY
	DECLARE @ResultM MONEY, @ResultD MONEY
	DECLARE @indicePeriodo INT
	DECLARE @inicioMes DATETIME

	SET @indicePeriodo = YEAR(@Fecha) * 100 + MONTH(@Fecha)
	SELECT @ResultM = COALESCE(SUM(Balance), 0)
		FROM BalancesMensuales 
		WHERE TipoEntidad = @TipoEntidad 
		AND CodigoEntidad = @CodigoEntidad 
		AND IndicePeriodo < @IndicePeriodo
	SET @inicioMes = DATEADD(d, -(DAY(@Fecha) - 1), @Fecha)
	SELECT @ResultD = COALESCE(SUM(Debe - Haber), 0)
			FROM DetallesCuentaEntidad
			WHERE CodigoEntidad = @CodigoEntidad 
			AND TipoEntidad = @TipoEntidad
			AND Fecha BETWEEN @inicioMes AND @Fecha
	SET @Result = @ResultM + @ResultD
	RETURN @Result
END
GO


-- ===================================================================================================================
-- Devuelve la existencia general de un articulo para una fecha dada:
CREATE FUNCTION dbo.ExistenciaItemFecha
(
	@CodigoItem nVarChar(20),
	@Fecha DATETIME
)
RETURNS FLOAT
AS
BEGIN
	DECLARE @Result FLOAT
	DECLARE @ResultM FLOAT, @ResultD FLOAT
	DECLARE @indicePeriodo INT
	DECLARE @inicioMes DATETIME

	SET @indicePeriodo = YEAR(@Fecha) * 100 + MONTH(@Fecha)
	SELECT @ResultM = COALESCE(SUM(Balance), 0)
		FROM BalancesMensuales 
		WHERE TipoEntidad = 'INV' 
		AND CodigoEntidad = @CodigoItem 
		AND IndicePeriodo < @IndicePeriodo
	SET @inicioMes = DATEADD(d, -(DAY(@Fecha) - 1), @Fecha)
	SELECT @ResultD = COALESCE(SUM(Entradas - Salidas), 0)
			FROM DetallesMinv 
			WHERE CodigoItem = @CodigoItem 
			AND FechaOperacion BETWEEN @inicioMes AND @Fecha
	SET @Result = @ResultM + @ResultD
	RETURN @Result
END
GO

-- ===================================================================================================================
-- Devuelve el valor de la existencia general de un articulo para una fecha dada, en funcion de los costos reales
-- de las transacciones registradas
CREATE FUNCTION dbo.ValorItemFecha
(
	@CodigoItem nVarChar(20),
	@Fecha DATETIME
)
RETURNS MONEY
AS
BEGIN
	DECLARE @Result MONEY
	DECLARE @ResultM MONEY, @ResultD MONEY
	DECLARE @indicePeriodo INT
	DECLARE @inicioMes DATETIME

	SET @indicePeriodo = YEAR(@Fecha) * 100 + MONTH(@Fecha)
	SELECT @ResultM = COALESCE(SUM(Balance), 0)
		FROM BalancesMensuales 
		WHERE TipoEntidad = 'VIV' 
		AND CodigoEntidad = @CodigoItem 
		AND IndicePeriodo < @IndicePeriodo
	SET @inicioMes = DATEADD(d, -(DAY(@Fecha) - 1), @Fecha)
	SELECT @ResultD = COALESCE(SUM((Entradas - Salidas) * Costo), 0)
			FROM DetallesMinv 
			WHERE CodigoItem = @CodigoItem 
			AND FechaOperacion BETWEEN @inicioMes AND @Fecha
	SET @Result = @ResultM + @ResultD
	RETURN @Result
END
GO


-- ===================================================================================================================
-- Devuelve una tabla con los costos acumulados de todos los productos al inicio de un mes
CREATE FUNCTION dbo.CostoAcumuladoProducto(@PrimerDiaMes DATETIME)
RETURNS @InventarioInicial TABLE
  (
	Codigo nvarchar(20),
	Descripcion nvarchar(80),
	Saldo money
  )
AS
BEGIN
	INSERT @InventarioInicial
		SELECT I.Codigo, I.Descripcion,SUM(Balance)
		FROM ItemsInventario I, BalancesMensuales D
		WHERE D.IndicePeriodo < Year(@PrimerDiaMes) * 100 + Month(@PrimerDiaMes)
			AND D.CodigoEntidad = I.Codigo
			AND D.TipoEntidad = 'VIV'
		GROUP BY I.Codigo, I.Descripcion
		ORDER BY I.Codigo
	RETURN
END
GO

-- ===================================================================================================================
-- Devuelve una tabla con la existencia de todos los productos al inicio de un mes
CREATE FUNCTION dbo.ExistenciaAcumuladaProducto(@PrimerDiaMes DATETIME)
RETURNS @InventarioInicial TABLE
  (
	Codigo nvarchar(20),
	Descripcion nvarchar(80),
	Saldo money
  )
AS
BEGIN
	INSERT @InventarioInicial
		SELECT I.Codigo, I.Descripcion,SUM(Balance)
		FROM ItemsInventario I, BalancesMensuales D
		WHERE D.IndicePeriodo < Year(@PrimerDiaMes) * 100 + Month(@PrimerDiaMes)
			AND D.CodigoEntidad = I.Codigo
			AND D.TipoEntidad = 'INV'
		GROUP BY I.Codigo, I.Descripcion
		ORDER BY I.Codigo
	RETURN
END
GO

-- ===================================================================================================================
-- Devuelve una tabla con los saldos de todas las entidades de un tipo a una fecha dadad
CREATE FUNCTION [dbo].[SaldoAcumulado](@TipoEntidad NVARCHAR(3), @Fecha DATETIME)
RETURNS @SaldosAcumulados TABLE
  (
	Codigo nvarchar(20),
	Saldo money
  )
AS
BEGIN
	INSERT @SaldosAcumulados
		SELECT Codigo, SUM(S) As Saldo FROM (
			SELECT CodigoEntidad As Codigo, SUM(Balance) As S
				FROM BalancesMensuales
				WHERE TipoEntidad = @TipoEntidad
				AND IndicePeriodo < YEAR(@fecha) * 100 + MONTH(@fecha)
				GROUP BY CodigoEntidad
			UNION SELECT
				CodigoEntidad, SUM(Debe - Haber)
				FROM DetallesCuentaEntidad
				WHERE Fecha BETWEEN DATEADD(d, -(DAY(@fecha) - 1), @fecha) AND @fecha
				AND TipoEntidad = @TipoEntidad
				GROUP BY CodigoEntidad) T
			GROUP BY Codigo
		ORDER BY Codigo
	RETURN
END
GO
